#!/usr/bin/perl -w
use Tk;
use Tk::BrowseEntry;
no warnings; # otherwise the script will return warnings while sorting non-numerical arrays numerical. Nevertheless, sorting is performed correctly.

### DEFAULT SETTINGS ###
$input_file_name="select your input file";
$next_neighbors=10;
$do_resampling="YES";
$resampling_steps=100;
$avoid_redundant_resamplings="YES";
$output_file_name="output.txt";

### GRAPHICAL USER INTERFACE ###
$mw=MainWindow->new(-title=>"spatial overlap assessor",-background=>White);
$mw->geometry("500x300"),
$mw->resizable(0,0);

$mw->Label(-text=>"Input file:",-background=>White)->place(-height=>20,-x=>30,-y=>30);
$filetypes=[['TEXT files','.txt'],['All Files','*']];
$entry_input_file=$mw->Entry(-width=>30,-textvariable=>\$input_file_name)->place(-height=>20,-x=>180,-y=>30);
$browse_input_file=$mw->Button(-text=>"browse",-command=>sub{$input_file_name=$mw->getOpenFile(-filetypes=>$filetypes,-defaultextension=>'.txt');$entry_input_file->configure(-textvariable=>$input_file_name);})->place(-width=>50,-height=>20,-x=>380,-y=>30);

$mw->Label(-text=>"Nearest neighbors:",-background=>White)->place(-height=>20,-x=>30,-y=>60);
$entry_next_neighbors=$mw->Entry(-width=>6,-textvariable=>\$next_neighbors,-justify=>'right')->place(-height=>20,-x=>180,-y=>60);

$mw->Label(-text=>"Perform resampling:",-background=>White)->place(-height=>20,-x=>30,-y=>90);
$ask_for_resampling=$mw->BrowseEntry(-variable=>\$do_resampling,-width=>4,-listwidth=>11,-listheight=>2,-state=>'readonly')->place(-x=>180,-y=>90);
$ask_for_resampling->insert('end',"YES","NO");

$mw->Label(-text=>"Resampling steps:",-background=>White)->place(-height=>20,-x=>30,-y=>120);
$entry_resampling_steps=$mw->Entry(-width=>6,-textvariable=>\$resampling_steps,-justify=>'right')->place(-height=>20,-x=>180,-y=>120);

$mw->Label(-text=>"Avoid redundant resamplings:",-background=>White)->place(-height=>20,-x=>30,-y=>150);
$ask_for_avoid_redundant_resamplings=$mw->BrowseEntry(-variable=>\$avoid_redundant_resamplings,-width=>4,-listwidth=>11,-listheight=>2,-state=>'readonly')->place(-x=>180,-y=>150);
$ask_for_avoid_redundant_resamplings->insert('end',"YES","NO");

$mw->Label(-text=>"Output file:",-background=>White)->place(-height=>20,-x=>30,-y=>180);
$entry_output_file=$mw->Entry(-width=>30,-textvariable=>\$output_file_name)->place(-height=>20,-x=>180,-y=>180);

$button_exit=$mw->Button(-text=>"EXIT",-command=>sub{exit;})->place(-width=>50,-height=>20,-x=>225,-y=>220);
$button_start=$mw->Button(-text=>"START",-command=>sub{check_settings();})->place(-width=>50,-height=>20,-x=>225,-y=>250);

MainLoop;

### START COMPUTATION ###
sub start
	{
	### disable widgets while computation ###
	$entry_input_file->configure(-state=>'disabled');
	$browse_input_file->configure(-state=>'disabled');
	$entry_next_neighbors->configure(-state=>'disabled');
	$ask_for_resampling->configure(-state=>'disabled');
	$entry_resampling_steps->configure(-state=>'disabled');
	$ask_for_avoid_redundant_resamplings->configure(-state=>'disabled');
	$entry_output_file->configure(-state=>'disabled');
	$button_exit->configure(-state=>'disabled');
	$button_start->configure(-state=>'disabled');

	### assign distances to arrays -> one array per input column ###
	@table=();
	$row_id=-1;
	foreach$row(@input)
		{
		$row_id++;
		chomp$row;
		@distances=split("\t",$row);
		$column_id=-1;
		foreach$distance(@distances)
			{
			$column_id++;
			$table[$column_id][$row_id]=$distance;
			}
		}

	### sort arrays and keep next n neighbors ###	
	if($do_resampling eq"YES")
		{push(@results,"Input file: $input_file_name\nIndividuals in group $species_A: $n_columns\nIndividuals in group $species_B: $group_B_individuals\nNumber of next neighbors to be analyzed: $next_neighbors\nResampling steps: $resampling_steps\nAvoid redundant resamplings: $avoid_redundant_resamplings\n\n");}
	else
		{push(@results,"Input file: $input_file_name\nIndividuals in group $species_A: $n_columns\nIndividuals in group $species_B: $group_B_individuals\nNumber of next neighbors to be analyzed: $next_neighbors\n\n");}
	sort_array();
	push(@results,"Average results original dataset:\t$species_B: $Mm\t$species_A: $Mg\tTotal average distance: $original_dataset_average_distance_average\tHighest distance average: $original_dataset_highest_distance_average\n\n\n");

	### resampling ###
	if($do_resampling eq"YES")
		{
		$individual_id=0;
		$Mm_for_total_average=0;
		$Mg_for_total_average=0;
		$highest_distance_total_average=0;
		$average_distance_total_average=0;
		foreach$reference(@table)
			{
			@subarray=@$reference;
			@random_assembly_array=();
			@majority=splice(@subarray,$n_columns);
			$found_different_combinations=0;
			while($found_different_combinations<$resampling_steps)
				{
				@random_assembly=();
				$randoms=0;
				while($randoms!=$n_columns)
					{
					$random_number=rand();
					$random_number=int($random_number*@majority);
					
					# avoid redundant numbers within random assembly
					unless($random_number~~@random_assembly)
						{
						push(@random_assembly,$random_number);
						$randoms++;
						}
					}
				@random_assembly=sort{$a<=>$b}@random_assembly;
				$random_assembly=join(' ',@random_assembly);
				
				# optional check for redundace of random assemblies
				if($avoid_redundant_resamplings eq"YES")
					{
					unless($random_assembly~~@random_assembly_array)
						{
						push(@random_assembly_array,$random_assembly);
						$found_different_combinations++;
						}
					}
				
				# ignore redundace of random assemblies
				elsif($avoid_redundant_resamplings eq"NO")
					{
					push(@random_assembly_array,$random_assembly);
					$found_different_combinations++;
					}
				}
			
			### assemble resampled arrays, send to sub ###
			$resampling_id=0;
			$individual_id++;
			push(@results,"\nresampling for individual $individual_id\n");
			$Mm_for_individual_average=0;
			$Mg_for_individual_average=0;
			$highest_distance_individual_average=0;
			$average_distance_individual_average=0;
			foreach$resampling(@random_assembly_array)
				{
				@restore=@subarray;
				$resampling_id++;
				@random_assembly=split(" ",$resampling);
				foreach(@random_assembly)
					{
					push(@subarray,$majority[$_]);	
					}
				@subarray=sort{$a<=>$b}@subarray;
				$Mm=0;
				$Mg=0;
				count();
				@subarray=@restore;
				push(@results,"resampling $resampling_id:\t$species_B: $Mm\t$species_A: $Mg\tAverage distance: $average_distance\tHighest distance: $highest_distance\n");
				$Mm_for_individual_average=$Mm_for_individual_average+$Mm;
				$Mg_for_individual_average=$Mg_for_individual_average+$Mg;
				$highest_distance_individual_average=$highest_distance_individual_average+$highest_distance;
				$average_distance_individual_average=$average_distance_individual_average+$average_distance;
				}
			$Mm_for_individual_average=$Mm_for_individual_average/@random_assembly_array;
			$Mg_for_individual_average=$Mg_for_individual_average/@random_assembly_array;
			$highest_distance_individual_average=$highest_distance_individual_average/@random_assembly_array;
			$average_distance_individual_average=$average_distance_individual_average/@random_assembly_array;
			push(@results,"Average for resampling individual $individual_id:\t$species_B: $Mm_for_individual_average\t$species_A: $Mg_for_individual_average\tAverage distance: $average_distance_individual_average\tHighest distance (average): $highest_distance_individual_average\n");
			$Mm_for_total_average=$Mm_for_total_average+$Mm_for_individual_average;
			$Mg_for_total_average=$Mg_for_total_average+$Mg_for_individual_average;
			$highest_distance_total_average=$highest_distance_total_average+$highest_distance_individual_average;
			$average_distance_total_average=$average_distance_total_average+$average_distance_individual_average;
			}
		$Mm_for_total_average=$Mm_for_total_average/$n_columns;
		$Mg_for_total_average=$Mg_for_total_average/$n_columns;
		$highest_distance_total_average=$highest_distance_total_average/$n_columns;
		$average_distance_total_average=$average_distance_total_average/$n_columns;
		push(@results,"\n\nTotal average for resampling $n_columns individuals:\t$species_B: $Mm_for_total_average\t$species_A: $Mg_for_total_average\tAverage distance: $average_distance_total_average\tHighest distance (average): $highest_distance_total_average\n");
		}

	### print results in output file ###
	open(OUTPUT,">$output_file_name");
	foreach(@results)
		{
		print OUTPUT"$_";
		}
	close OUTPUT;
	@results=();
	$mw->messageBox(-message=>"Results are saved at $output_file_name",-type=>'ok',-icon=>'info');
	
	### enable widgets after computation ###
	$entry_input_file->configure(-state=>'normal');
	$browse_input_file->configure(-state=>'normal');
	$entry_next_neighbors->configure(-state=>'normal');
	$ask_for_resampling->configure(-state=>'normal');
	$entry_resampling_steps->configure(-state=>'normal');
	$ask_for_avoid_redundant_resamplings->configure(-state=>'normal');
	$entry_output_file->configure(-state=>'normal');
	$button_exit->configure(-state=>'normal');
	$button_start->configure(-state=>'normal');
	}

### subroutines ###
sub sort_array
	{
	$Mm=0;
	$Mg=0;
	$original_dataset_highest_distance_average=0;
	$original_dataset_average_distance_average=0;
	$original_individual_ID=0;
	foreach$reference(@table)
		{
		$original_individual_ID++;
		$Mm_original=0;
		$Mg_original=0;
		@subarray=@$reference;
		@subarray=sort{$a<=>$b}@subarray;
		count();
		sub count
			{
			@next_neighbors=();
			$elements=-1;
			foreach(0..@subarray-1)
				{
				$elements++;
				if($elements<=$next_neighbors||substr($subarray[$next_neighbors-1],0,4)eq substr($subarray[$_],0,4))
					{
					push(@next_neighbors,$subarray[$_]);
					}
				}
			# count next neighbors (pro rata) #
			$closing_equalities=0;
			$highest_distance=substr($next_neighbors[-1],0,4);
			foreach(@next_neighbors)
				{
				if(substr($_,0,4)eq$highest_distance)
					{
					$closing_equalities++;
					}
				}
			$available_places=$next_neighbors-((@next_neighbors-1)-$closing_equalities);
			foreach(1..(@next_neighbors-1)-$closing_equalities)
				{
				if(substr($next_neighbors[$_],-2)eq"$species_B")
					{$Mm=$Mm+1;
					$Mm_original=$Mm_original+1;}
				elsif(substr($next_neighbors[$_],-2)eq"$species_A")
					{$Mg=$Mg+1;
					$Mg_original=$Mg_original+1;}
				}
			foreach(((@next_neighbors-1)-$closing_equalities)+1..@next_neighbors-1)
				{
				if(substr($next_neighbors[$_],-2)eq"$species_B")
					{$Mm=$Mm+($available_places/$closing_equalities);
					$Mm_original=$Mm_original+($available_places/$closing_equalities);}
				if(substr($next_neighbors[$_],-2)eq"$species_A")
					{$Mg=$Mg+($available_places/$closing_equalities);
					$Mg_original=$Mg_original+($available_places/$closing_equalities);}
				}
			# calculate average distance from individual 0
			$average_distance=0;
			foreach(1..$next_neighbors)
				{
				$distance=substr($next_neighbors[$_],0,4);
				$average_distance=$average_distance+$distance;
				}
			$average_distance=$average_distance/$next_neighbors;
			}
		$original_dataset_highest_distance_average=$original_dataset_highest_distance_average+$highest_distance;
		$original_dataset_average_distance_average=$original_dataset_average_distance_average+$average_distance;
		push(@results,"Results individual $original_individual_ID:\t$species_B: $Mm_original\t$species_A: $Mg_original\tAverage distance: $average_distance\tHighest distance: $highest_distance\n");
		}
	$Mm=$Mm/@table;
	$Mg=$Mg/@table;
	$original_dataset_average_distance_average=$original_dataset_average_distance_average/@table;
	$original_dataset_highest_distance_average=$original_dataset_highest_distance_average/@table;
	}

sub check_settings
	{
	### read input file ###
	$error_status=0;
	$could_not_open=0;
	$error_message="Error in settings:\n";
	
	unless(open(TEST_OPEN,"$input_file_name"))
		{
		$error_status=1;
		$could_not_open=1;
		$error_message.="Could not open input file!\n";
		}
	else
		{
		close TEST_OPEN;	
		}
	open(INPUT,"$input_file_name");
	@input=<INPUT>;
	close INPUT;
	$n_columns=($input[0]=~tr/\t//);
	$n_columns++; 						# individuals group A
	$group_B_individuals=@input-$n_columns;	# individuals group B

	### get species abbreviations ###
	$species_A=substr($input[0],-3,2);
	$species_B=substr($input[-1],-2,2);

	### check consistent use of abbreviations ###
	$total_distances=@input*$n_columns;
	$input_as_string=join("\t",@input);
	$input_as_string=~s/\n//g;
	$input_as_string.="\t";
	$abbreviation_A=($input_as_string=~s/$species_A\t//g);
	$abbreviation_B=($input_as_string=~s/$species_B\t//g);
	if($total_distances!=$abbreviation_A+$abbreviation_B)
		{
		$error_status=1;
		$error_message.="Inconsistent use of group name abbreviations in input file!\n";
		}

	### check user input ###
	if($species_A eq $species_B)
		{
		$error_status=1;
		$error_message.="Group names have to be different!\n";
		}
	if($next_neighbors!~/^\d+$/||$next_neighbors==0||$next_neighbors>@input)
		{
		$error_status=1;
		$error_message.="Number of nearest neighbors must be an integer >0 and <= number of individuals in input file!\n";
		}
	if($resampling_steps!~/^\d+$/||$resampling_steps==0)
		{
		$error_status=1;	
		$error_message.="Resampling steps must be an integer >0!\n";
		}
	if($output_file_name eq""||$output_file_name=~/[\\\/:\*\?><"|]/)
		{
		$error_status=1;	
		$error_message.="Invalid output file name!\n";
		}
	if(@input<=$n_columns*2&&$do_resampling eq"YES"&&$could_not_open==0)
		{
		$error_status=1;	
		$error_message.="Resampling not possible. Improper proportion of individuals!\n";
		}

	### calculate number of nonredundant resampling possibilities ###
	$n_excess=@input-$n_columns;
	$excess=$n_excess-$n_columns;
	$resampling_possibilities=1;
	foreach(1..$excess)
		{
		$resampling_possibilities=$resampling_possibilities*$n_excess;
		$n_excess=$n_excess-1;
		}
	$resampling_possibilities=$resampling_possibilities/2;
	
	if($do_resampling eq"YES"&&$avoid_redundant_resamplings eq"YES"&&$resampling_possibilities<$resampling_steps&&$could_not_open==0)
		{
		$error_status=1;	
		$error_message.="Number of nonredundant resampling steps exceeds number of nonredundant resampling possibilities ($resampling_possibilities)!\n";
		}

	### start computation ###
	if($error_status==0)
		{
		start();
		}

	### return erros message ###
	elsif($error_status==1)
		{
		$mw->messageBox(-message=>"$error_message",-type=>'ok',-icon=>'error');
		}
	}